-------------------------------------------------------------------------------
--
--		cranky - "RUNNING"
--		Bubble Tea Invitational 2.5
--		
--		Author: 	Telperion
--		Date: 		2017-05-16
--		Target:		SM5.0.12+
--
-------------------------------------------------------------------------------
--
--		you can't keep running running running running running running
--		you can't keep running running running running runnin' runnin'
--		you can't keep running, running, running away, running away
--
-------------------------------------------------------------------------------

local circumvention = false
if circumvention then
	return Def.ActorFrame {}
end

local sw = SCREEN_WIDTH
local sh = SCREEN_HEIGHT
local BPS = GAMESTATE:GetSongBPS()
local ofs = 0.009
local overtime = 0
local visualOffset = -0.01
local fgmsg = 0
local checked = false
local plr = {nil, nil}

local PI = math.pi
local LOG2 = math.log(2.0)
local DEG_TO_RAD = math.pi / 180.0

--
-- 		some funktion !
--
function SideSign(i) return (i == 2) and 1 or -1 end

function RangeScale(t, inLower, inUpper, outLower, outUpper)
	local ti = (t - inLower) / (inUpper - inLower)
	return outLower + ti * (outUpper - outLower)
end

function RangeClamp(t, outLower, outUpper)
	local ti = t
	ti = (ti < outLower) and outLower or ti
	ti = (ti > outUpper) and outUpper or ti
	return ti
end

function Whomst(act)
	return tonumber(string.match(act:GetName(), "P([0-9]+)"))
end

function Whichst(act)
	return tonumber(string.match(act:GetName(), "_([0-9]+)"))
end

function Wherest(act)
	return tonumber(string.match(act:GetName(), "Q([0-9]+)"))
end

function GentleFunction(t, p)
	-- Quintic tweening function for smooth behavior
	-- p = control parameter determining slope in center
	-- p el [  0, 15/8] for monotonic movement
	-- p el [5/4,  3/2] for non-inflective movement
    local tt = 2*t - 1
    local z = (  p - 1.5) * tt^5
    		+ (2.5 - 2*p) * tt^3
    		+ 		   p  * tt
    return 0.5*z + 0.5
end

function CuteFunction(t)
	-- Smoothest way to get from point A to B on a quintic
	-- Specific case of GentleFunction with minimal inflection (p = 5/4)
    local tt = 2*t - 1
    local z = -0.25 * tt^5
    		+  1.25 * tt
    return 0.5*z + 0.5
end

function StrongFunction(t, p)
	-- Cubic tweening function for smooth behavior
	-- p = control parameter determining slope at ends
	-- p el [0, 3] for monotonic movement
    local tt = 2*t - 1
    local z = (p - 2) * tt^3 
    		+ (3 - p) * tt
    return 0.5*z + 0.5
end

function BiasFunction(t, p, q)
	-- Cubic tweening function for smooth behavior with off-center middle
	-- p = control parameter determining slope at middle
	-- q = control parameter determining y-intercept (and interest point)
    local tt = 2*t - 1
    local z = (((1 - p)  * tt
    		+   (0 - q)) * tt
    		+        p ) * tt
    		+        q
    return 0.5*z + 0.5
end

function SortModsTable(mt)
	-- Insertion sort into the butt
	-- because wow I'm mostly lazy
	-- and can assume it's mostly sorted

	mtNew = {}

	for _,mr in pairs(mt) do
		local insertLocation = 1
		for i = #mtNew,1,-1 do
--			Trace("### SortModsTable: "..mtNew[i][1].." vs. "..mr[1])
			if mtNew[i][1] <= mr[1] then
				insertLocation = i+1
				break
			end
		end
--		Trace("### SortModsTable: "..mr[1].." ("..mr[2]..") -> "..insertLocation)
		table.insert(mtNew, insertLocation, mr)
	end

	return mtNew
end


local _DZ_ = Def.ActorFrame {};

local _FG_ = Def.ActorFrame {
	InitCommand = function(self)
		self:fov(45)
			:vanishpoint(sw/2, sh/4)			
			:SetDrawByZPosition(true)
	end,
	OnCommand = function(self)
		plr[1] = SCREENMAN:GetTopScreen():GetChild('PlayerP1')
		plr[2] = SCREENMAN:GetTopScreen():GetChild('PlayerP2')
		SCREENMAN:GetTopScreen():SetDrawByZPosition(true)


		local hamburger = SCREENMAN:GetTopScreen()
		if hamburger:GetScreenType() == "ScreenType_Gameplay" then
			hamburger:GetChild("Overlay" ):decelerate(4.0 / BPS):diffusealpha(0.0)
			hamburger:GetChild("Underlay"):decelerate(4.0 / BPS):diffusealpha(0.0)
		end
				
		local playersFound = 0
		for pn = 1,2 do
			pv = plr[pn]
			if pv then
				pv:GetChild("Combo"):hibernate(1573)
			end
			
			pops = GAMESTATE:GetPlayerState("PlayerNumber_P"..pn):GetPlayerOptions("ModsLevel_Song")
			pops:FailSetting('FailType_Off')
		end

		self:sleep(1573)
	end
};



-------------------------------------------------------------------------------
--
-- 		Proxies (as usual)
--

--
--	Player proxies (one apiece)
--
local PlayerProxies = {nil, nil}
for pn = 1,2 do
	i = 1
	_FG_[#_FG_ + 1] = Def.ActorFrame {	
		Name = "ProxyP"..pn.."Outer",
		Def.ActorFrame {	
			Name = "ProxyP"..pn.."Inner",
			Def.ActorProxy {					
				Name = "ProxyP"..pn,
				InitCommand = function(self)
					self:aux( Whomst(self) )
				end,
				BeginCommand=function(self)
					local McCoy = SCREENMAN:GetTopScreen():GetChild('PlayerP'..self:getaux())
					if McCoy then self:SetTarget(McCoy) else self:hibernate(1573) end
				end,
				OnCommand=function(self)
					local McCoy = SCREENMAN:GetTopScreen():GetChild('PlayerP'..self:getaux())
					if McCoy then 
						self:xy(-McCoy:GetX(), -McCoy:GetY())
						self:GetParent():xy(McCoy:GetX(), McCoy:GetY())
					end
				end,
				RecenterProxyMessageCommand=function(self)					
					local McCoy = SCREENMAN:GetTopScreen():GetChild('PlayerP'..self:getaux())
					if McCoy then 
						self:xy(-McCoy:GetX(), -McCoy:GetY())
						self:GetParent():xy(McCoy:GetX(), McCoy:GetY())
					end
				end
			},
			InitCommand = function(self)
--				self:aux( Whichst(self) )
			end,
			OnCommand = function(self)
			end,
		},
		InitCommand = function(self)
--			self:aux( Whichst(self) )
			PlayerProxies[pn] = self
		end,
		OnCommand = function(self)
			self:xy(sw * 0.25 * SideSign(Whomst(self)), 0)
				:z(1)
		end,
	}
end



--
--	Column proxies 
--
local ttw = math.pow(2, math.ceil(math.log(sw) / LOG2))
local tth = math.pow(2, math.ceil(math.log(sh) / LOG2))
local twscale = sw / ttw
local thscale = sh / tth
local laneSeparationParameter = 2.0
local laneCroppingParameter = 0.5 * (sh/sw * 4.0/3.0)		-- screen resolution plays with this parameter very differently
local laneMinifyingParameter = 0.7
local columnProxyCount = 3				-- this many, for each player

local act_columns = {}
for proxyIndex = 1,2*columnProxyCount do
	act_columns[#act_columns + 1] = {nil, nil, nil, nil}
end

local runnerSimonActor	= nil
local runnerSimonSpeed	= 1
local runnerSomeReverse = false
local runnerGoalpost	= 1573
local runnerPeriod 		= 12.0
local runnerOnScreen 	= 0.5
local runnerKnee 		= 2.0 / runnerPeriod
local runnerStand 		= 6.0 / runnerPeriod
local runnerOffset 		= -runnerKnee*1.5 + 0.5
local runnerAnchor 		= math.ceil(columnProxyCount / 2)

function DivergenceDude(t, t0, t1, x0, m)
	if t <= t0 then
		-- Decelerating quadratic approach to (t0, x0).
		local tt = (t / t0 - 1)
		return x0 * (1 - tt*tt)
	elseif t <= t1 then
		-- Accelerating quadratic approach from (t0, x0) until tangent line has slope of (m).
		local tt = (t - t0)
		return (m*tt*tt) / (2*t1*(t1 - t0)) + x0
	else
		-- Linear escape with slope (m).
		return m*(t - t1) + m*(t1 - t0)/(2*t1) + x0
	end
end

function LopsidedAlpha(t, knee)
	if t <= knee then
		return t / knee
	else
		return 1 - (t - knee) / (1 - knee)
	end
end

for pn = 1,2 do
	for laneIndex = 1,4 do
		_FG_[#_FG_ + 1] = Def.ActorFrameTexture {
			Name = "TextureP"..pn.."ForLane_"..laneIndex,
			InitCommand = function(self)
				self:SetTextureName( self:GetName() )
					:SetWidth(sw)
					:SetHeight(sh)
					:EnableAlphaBuffer( true )
					:Create()				
			end,
--			Def.Quad {
--				InitCommand = function(self)
--					self:xy(sw * 0.5, sh * 0.5)
--						:SetWidth(sw)
--						:SetHeight(sh)
--						:diffuse({0.8, 0.8, 1.0, 0.2})
--				end
--			},
			Def.ActorProxy {
				Name = "LaneProxyP"..pn.."_"..laneIndex,
				BeginCommand = function(self)
					local McCoy = SCREENMAN:GetTopScreen()
										:GetChild('PlayerP'..Whomst(self))
					if McCoy then self:SetTarget(McCoy) else self:hibernate(1573) end
				end,
				RecenterProxyMessageCommand = function(self)
					self:x((2.5 - Whichst(self)) * (64 + 128 * laneSeparationParameter) * laneMinifyingParameter)
				end,
			}
		}


		for copyIndex = 0,columnProxyCount do 
			-- Yes, make one extra of each player! Use it for ~other~ purposes.
			_FG_[#_FG_ + 1] = Def.ActorFrame {	
				Name = "OuterLaneP"..pn.."_"..laneIndex.."Q"..copyIndex,
				Def.Sprite {
					Name = "InnerLaneP"..pn.."_"..laneIndex.."Q"..copyIndex,
					InitCommand = function(self)
						act_columns[copyIndex+1 + (pn-1)*columnProxyCount][laneIndex] = self

						self:SetTexture(
							self:GetParent():GetParent()
								:GetChild("TextureP"..Whomst(self).."ForLane_"..Whichst(self))
								:GetTexture()
							)

						self:customtexturerect(
							    0.5 * (1.0 - laneCroppingParameter) * twscale,
							    0.0,
								0.5 * (1.0 + laneCroppingParameter) * twscale,
								1.0
							)
							:zoomx(laneCroppingParameter)
							:zoomy(1.0 / thscale)
					end
				},
				InitCommand = function(self)
					self:xy(sw * 0.5 + 64 * (Whichst(self) - 2.5) * laneMinifyingParameter, sh * 0.5)
						:zoom(1.0)
						:diffusealpha(0)
				end,
				ShowColumnProxiesMessageCommand = function(self, args)
					local tweenTime = args[2] and args[2] or 0
					
					for _,cpi in pairs(args[1]) do
						if Wherest(self) == cpi - 1 then
							self:visible(true)
								:decelerate(tweenTime / BPS)
								:diffusealpha(1)
						end
					end
				end,
				HideColumnProxiesMessageCommand = function(self, args)
					local tweenTime = args[2] and args[2] or 0

					for _,cpi in pairs(args[1]) do
						if Wherest(self) == cpi - 1 then
							self:accelerate(tweenTime / BPS)
								:diffusealpha(0)
								:queuecommand("HideThisColumnProxy")
						end
					end
				end,
				HideThisColumnProxyCommand = function(self)
					self:visible(false)
				end,
			}
		end
	end
end


--		ListChildrenMessageCommand = function(self)
--			local pp = SCREENMAN:GetTopScreen():GetChild('PlayerP'..self:getaux())
--			local nf = pp:GetChild('NoteField')
--
--			pp:decelerate(4.0 / BPS)
--			  :xy(0, 0)
--			for i,cv in ipairs(nf:GetColumnActors()) do
--				signA = (i > 2) and 1 or -1
--				signB = (i % 2) and 1 or -1
--
--				cv:GetPosHandler():SetSplineMode('NoteColumnSplineMode_Position')
--				cv:GetPosHandler():SetBeatsPerT(1.3 - i * 0.2)
--				local splp = cv:GetPosHandler():GetSpline()
--				splp:SetSize(4)
--				splp:SetPoint(1, {sw *  0.0		  + i*64-160, sh *  0.2, 0})
--				splp:SetPoint(2, {sw * -0.1 * signA + i*64-160, sh *  0.5, 0})
--				splp:SetPoint(3, {sw *  0.1 * signA + i*64-160, sh *  0.8, 0})
--				splp:SetPoint(4, {sw *  0.0		  + i*64-160, sh *  1.1, 0})
--				splp:Solve()
--
--				cv:GetZoomHandler():SetSplineMode('NoteColumnSplineMode_Position')
--				cv:GetZoomHandler():SetBeatsPerT(0.7 + i * 0.2)
--				local splz = cv:GetZoomHandler():GetSpline()
--				splz:SetSize(5)
--				splz:SetPoint(1, {1.0, 1.0, 1.0})
--				splz:SetPoint(2, {0.3, 1.0, 1.0})
--				splz:SetPoint(3, {1.0, 1.0, 1.0})
--				splz:SetPoint(4, {1.0, 0.3, 1.0})
--				splz:SetPoint(5, {1.0, 1.0, 1.0})
--				splz:Solve()
--
--				cv:decelerate(4.0 / BPS)
--				  :x(0.5 * sw)
----				  :rotationz(20*i - 50)
--			end
--		end


--
--	Judgment proxies
--
for pn = 1,2 do
	_FG_[#_FG_ + 1] = Def.ActorProxy {
		Name = "JudgeP"..pn.."Proxy",
		InitCommand = function(self)
			self:aux( Whomst(self) )
		end,
		BeginCommand = function(self)
			local McCoy = SCREENMAN:GetTopScreen():GetChild('PlayerP'..self:getaux())
			if McCoy then 
				local McJudge = McCoy:GetChild('Judgment')
				self:SetTarget(McJudge)
				McJudge:visible(false)
			else 
				self:hibernate(1573)
			end
		end,
		OnCommand = function(self)
			local p = self:GetParent()
						  :GetChild('ProxyP'..self:getaux().."Outer")
						  :GetChild('ProxyP'..self:getaux().."Inner")
			
			self:xy(p:GetX(), sh/2)
		end,
	}
end

--
-- 		Proxies (as usual)
--
-------------------------------------------------------------------------------


-------------------------------------------------------------------------------
--
-- 		This is where the shit will be happening.
--

_FG_[#_FG_ + 1] = Def.Actor {
	InitCommand = function(self)
		runnerSimonActor = self
		self:aux(runnerOffset - runnerKnee * runnerPeriod * runnerOnScreen)
	end,

	SimonMarchMessageCommand = function(self, args)
		self:stoptweening()
		local prev 			= self:getaux()
		local tweenTime 	= args[1] and args[1] or 4
		local newSpeed		= args[2] and args[2] or runnerSimonSpeed
		runnerSimonSpeed	= newSpeed

		self:accelerate(args[1] / BPS)
			:aux(prev + (tweenTime * newSpeed) / 2)
			:queuecommand("SimonKeepMarching")
	end,
	SimonKeepMarchingCommand = function(self)
		local prev 		= self:getaux()
		self:linear(runnerGoalpost / BPS)
			:aux(prev + runnerGoalpost * runnerSimonSpeed)
	end,

	SimonHaltMessageCommand = function(self, args)
		self:stoptweening()
		local prev 			= self:getaux()
		local tweenTime 	= args[1] and args[1] or 4

		self:decelerate(args[1] / BPS)
			:aux(prev + (tweenTime * runnerSimonSpeed) / 2)
	end,
}


local messageList = {
	{  0.00, "HideColumnProxies", 	{{1,2,3}}},
	{  4.00, "RecenterProxy"},	

	{ 16.00, "HidePlayerProxies", 	{}},
	{ 16.00, "ShowColumnProxies", 	{{3}, 1}},

	{ 24.00, "ShowColumnProxies", 	{{1,2}, 0}},

	{ 30.00, "SimonMarch",			{2}},

	{ 30.00, "SimonMarch",			{2}},
	{ 60.00, "SimonHalt",			{3}},
	{ 63.00, "SimonMarch",			{1}},

	{216.00, "SimonHalt",			{6}},
	{222.00, "SimonMarch",			{2, 2}},

	{280.00, "SimonHalt",			{4}},
	{284.00, "RunnerReverse",		{3}},
	{287.00, "SimonMarch",			{1, -1}},

	{316.00, "SimonHalt",			{3}},
	{316.00, "RunnerNormal",		{3}},
	{319.00, "SimonMarch",			{1, 1}},

	{508.00, "SimonHalt",			{9}},
};

messageList = SortModsTable(messageList)
gfxUpdateFunction = function()
	-- Most things are determined by beat, believe it or not.		
	overtime = GAMESTATE:GetSongBeat() - visualOffset
	
	-- TODO: this assumes the effect applies over a constant BPM section!!
	BPS = GAMESTATE:GetSongBPS()
	
	-- Who's interesting today?
	if overtime >=   0.0 and not checked then
		for i,v in ipairs(plr) do
			if v then
				v:visible(false):decelerate(4.0 / BPS)
				 :x(sw/2)
--				 :y(sh/2 - 30)
				 :z(0)
			end
		end
		
		checked = true
	end
			
	-- Broadcast messages on their own terms.
	while true do
		if fgmsg < #messageList then
			messageBeat, messageName, messageArgs = unpack(messageList[fgmsg+1])
			if overtime >= messageBeat then			
				if messageArgs then
					MESSAGEMAN:Broadcast( messageName, messageArgs )
				else
					MESSAGEMAN:Broadcast( messageName )
				end
				
				fgmsg = fgmsg + 1
			else
				break
			end
		else
			break
		end
	end


	-- Update column actor positions from here.	
	local temp
	local runningTime = runnerSimonActor:getaux()
	local theirTime
	for copyIndex = 0,columnProxyCount-1 do
		for pn0 = 0,1 do
			local allCopyTimer = copyIndex * 2 + pn0 
			local allCopyIndex = pn0 * columnProxyCount + (copyIndex + pn0 * runnerAnchor) % columnProxyCount + 1

			theirTime = math.fmod( (runningTime - runnerOffset) / runnerPeriod 
								  + allCopyTimer / (2*columnProxyCount), 
								  1.0)
			theirTime = RangeClamp(theirTime / runnerOnScreen, 0.0, 1.0)
			local zoomTime = 1.0 + 0.6*theirTime

			for laneIndex = 1,4 do
				local magnetism = (laneIndex == 1 or laneIndex == 2) and -1 or 1

				local they = act_columns[allCopyIndex][laneIndex]
				if they then
					local theirTX = sw * 0.5 + DivergenceDude(theirTime, runnerKnee, runnerStand, zoomTime * 64 * (laneIndex - 2.5) * laneMinifyingParameter, 0.25 * sw * magnetism / runnerStand)
					local theirTY = sh * 0.5 -- sh * (0.5 + 0.2	* ((pn0 == 1 and runnerSomeReverse) and 1 or -1) * theirTime)
					theirParent = they:GetParent()
					theirParent:xy(theirTX, theirTY)
					theirParent:zoom(zoomTime)
					theirParent:diffusealpha(LopsidedAlpha(theirTime, runnerKnee))
				end
			end
		end
	end

--	Trace('Update @ ' .. GAMESTATE:GetSongBeat() .. " (" .. GAMESTATE:GetCurMusicSeconds() .. " sec.)")
end


-- Message handler for orphans
_FG_[#_FG_ + 1] = Def.ActorFrame {
	HidePlayerProxiesMessageCommand = function(self, args)
		local tweenTime = args[1] and args[1] or 2

		for i,v in ipairs(PlayerProxies) do
			if v then
				v:decelerate(tweenTime / BPS)
			 	 :y(sh * 1.0)
			 	 :diffusealpha(0)
			end
		end
	end,

	RunnerReverseMessageCommand = function(self, args)
		runnerSomeReverse = true
	end,
	RunnerNormalMessageCommand = function(self, args)
		runnerSomeReverse = false
	end,
}



_FG_[#_FG_ + 1] = Def.ActorFrame {
	InitCommand = function(self)
		self:SetUpdateFunction(gfxUpdateFunction)
	end,

	Def.ActorFrame {
		InitCommand = function(self)
			self:sleep(69420)
		end
	}
}


-- Load the mods table parser into this script.
niceSpeed = (420 + 69) / 145			-- This song is 145 BPM.
modsTable = {
	-- [1]: beat start
	-- [2]: mod type
	-- [3]: mod strength (out of unity),
	-- [4]: mod approach (in beats to complete)
	-- [5]: player application (1 = P1, 2 = P2, 3 = both, 0 = neither)
		
		{   0.000,	"ScrollSpeed",	niceSpeed,    8.0,	3}, 
		{   0.000,	"Dark",				  0.2,    8.0,	3}, 
		{   0.000,	"Centered",			 -0.1,    0.0,	3}, 

		{  16.000,	"Split",			  0.0,    2.0,	3}, 
		{  16.000,	"Reverse",			  0.0,    2.0,	3}, 
		{  16.000,	"Confusion",		  0.5,    0.0,	1}, 
		{  16.000,	"Confusion",		 -0.5,    0.0,	2}, 
		{  16.000,	"Dark",				  1.0,    2.0,	3}, 
		{  16.000,	"Flip",				 -laneSeparationParameter,    2.0,	3},
		{  16.000,	"Mini",			 2*(1-laneMinifyingParameter),    2.0,	3},
--		{  16.0,	"Tipsy",			  0.1,   16.0,	1},
		{  16.050,	"Confusion",		  0.0,    2.0,	3}, 
		{  18.000,	"Dark",				  0.0,    2.0,	3}, 
		{  23.900,	"Flip",			 -0.9*laneSeparationParameter,    0.1,	3},
		{  24.000,	"Flip",				 -laneSeparationParameter,    1.0,	3},


		{  64.000,	"Drunk",			  0.5,   28.0,	1}, 
		{  64.000,	"Drunk",			 -0.5,   28.0,	2}, 
		{  92.000,	"Drunk",			  0.0,    4.0,	3}, 

		{ 158.000,	"Drunk",			 -0.5,    2.0,	1}, 
		{ 158.000,	"Drunk",			  0.5,    2.0,	2}, 
		{ 160.000,	"Cross",			  0.0,    2.0,	3}, 
		{ 160.000,	"Reverse",			  0.0,    2.0,	3}, 
		{ 192.000,	"Drunk",			  0.0,   32.0,	3}, 

		{ 256.000,	"Wave",				  1.5,   60.0,	3}, 
		{ 256.000,	"Stealth",			  0.75,  24.0,	3}, 
		{ 280.000,	"Stealth",			  0.0,    4.0,	3}, 
		{ 284.000,	"Reverse",			  1.0,    4.0,	2}, 
		{ 316.000,	"Wave",				  0.0,    4.0,	3}, 
		{ 316.000,	"Reverse",			  0.0,    2.0,	2}, 

		{ 319.500,	"Beat",				 -0.5,    0.0,	1}, 
		{ 319.500,	"Beat",				  0.5,    0.0,	2}, 
		{ 348.000,	"Beat",				  0.0,    4.0,	3},  

		{ 380.000,	"Tipsy",			  0.2,    4.0,	1},
		{ 380.000,	"Tipsy",			 -0.2,    4.0,	2},
		{ 384.000,	"Tipsy",			  0.5,   16.0,	1},
		{ 384.000,	"Tipsy",			 -0.5,   16.0,	2},
		{ 404.000,	"Tipsy",			  0.0,    8.0,	3},

		{ 412.000,	"Drunk",			  1.0,   96.0,	1}, 
		{ 412.000,	"Drunk",			 -1.0,   96.0,	2}, 
		{ 507.500,	"Drunk",			  0.0,    4.0,	3}, 
}

for beatIndex = 2,14,2 do
	modStrength = RangeScale(beatIndex, 0, 16, 0, 0.1)
	modsTable[#modsTable + 1] = { beatIndex - 0.25,	"Split",	  modStrength*2,  0.5,	1 }
	modsTable[#modsTable + 1] = { beatIndex - 0.25,	"Reverse",	 -modStrength,    0.5,	1 } 
	modsTable[#modsTable + 1] = { beatIndex - 0.25,	"Split",	 -modStrength*2,  0.5,  2 } 
	modsTable[#modsTable + 1] = { beatIndex - 0.25,	"Reverse",	  modStrength,    0.5,	2 } 
	modsTable[#modsTable + 1] = { beatIndex + 0.75,	"Split",	 -modStrength*2,  0.5,  1 } 
	modsTable[#modsTable + 1] = { beatIndex + 0.75,	"Reverse",	  modStrength,    0.5,	1 } 
	modsTable[#modsTable + 1] = { beatIndex + 0.75,	"Split",	  modStrength*2,  0.5,	2 } 
	modsTable[#modsTable + 1] = { beatIndex + 0.75,	"Reverse",	 -modStrength,    0.5,	2 } 
end


for beatIndex = 96,152,8 do
	modStrength = RangeScale(beatIndex, 96, 160, 0.05, 0.15)
	modsTable[#modsTable + 1] = { beatIndex - 0.10,	"Cross",	  modStrength*2,  0.2,	1 }
	modsTable[#modsTable + 1] = { beatIndex - 0.10,	"Reverse",	 -modStrength,    0.2,	1 } 
	modsTable[#modsTable + 1] = { beatIndex - 0.10,	"Cross",	 -modStrength*2,  0.2,  2 } 
	modsTable[#modsTable + 1] = { beatIndex - 0.10,	"Reverse",	  modStrength,    0.2,	2 } 
	modsTable[#modsTable + 1] = { beatIndex + 0.65,	"Cross",	 -modStrength*2,  0.2,  1 } 
	modsTable[#modsTable + 1] = { beatIndex + 0.65,	"Reverse",	  modStrength,    0.2,	1 } 
	modsTable[#modsTable + 1] = { beatIndex + 0.65,	"Cross",	  modStrength*2,  0.2,	2 } 
	modsTable[#modsTable + 1] = { beatIndex + 0.65,	"Reverse",	 -modStrength,    0.2,	2 } 
	modsTable[#modsTable + 1] = { beatIndex + 1.40,	"Cross",	  0.0,			  0.2,  3 } 
	modsTable[#modsTable + 1] = { beatIndex + 1.40,	"Reverse",	  0.0,			  0.2,	3 } 

	modsTable[#modsTable + 1] = { beatIndex + 3.40,	"Cross",	  modStrength*2,  0.2,	2 }
	modsTable[#modsTable + 1] = { beatIndex + 3.40,	"Reverse",	 -modStrength,    0.2,	2 } 
	modsTable[#modsTable + 1] = { beatIndex + 3.40,	"Cross",	 -modStrength*2,  0.2,  1 } 
	modsTable[#modsTable + 1] = { beatIndex + 3.40,	"Reverse",	  modStrength,    0.2,	1 } 
	modsTable[#modsTable + 1] = { beatIndex + 3.65,	"Cross",	 -modStrength*2,  0.2,  2 } 
	modsTable[#modsTable + 1] = { beatIndex + 3.65,	"Reverse",	  modStrength,    0.2,	2 } 
	modsTable[#modsTable + 1] = { beatIndex + 3.65,	"Cross",	  modStrength*2,  0.2,	1 } 
	modsTable[#modsTable + 1] = { beatIndex + 3.65,	"Reverse",	 -modStrength,    0.2,	1 } 

	modsTable[#modsTable + 1] = { beatIndex + 4.10,	"Cross",	  modStrength*2,  0.2,	2 }
	modsTable[#modsTable + 1] = { beatIndex + 4.10,	"Reverse",	 -modStrength,    0.2,	2 } 
	modsTable[#modsTable + 1] = { beatIndex + 4.10,	"Cross",	 -modStrength*2,  0.2,  1 } 
	modsTable[#modsTable + 1] = { beatIndex + 4.10,	"Reverse",	  modStrength,    0.2,	1 } 
	modsTable[#modsTable + 1] = { beatIndex + 4.65,	"Cross",	 -modStrength*2,  0.2,  2 } 
	modsTable[#modsTable + 1] = { beatIndex + 4.65,	"Reverse",	  modStrength,    0.2,	2 } 
	modsTable[#modsTable + 1] = { beatIndex + 4.65,	"Cross",	  modStrength*2,  0.2,	1 } 
	modsTable[#modsTable + 1] = { beatIndex + 4.65,	"Reverse",	 -modStrength,    0.2,	1 } 
	modsTable[#modsTable + 1] = { beatIndex + 5.40,	"Cross",	  0.0,			  0.2,  3 } 
	modsTable[#modsTable + 1] = { beatIndex + 5.40,	"Reverse",	  0.0,			  0.2,	3 } 

	modsTable[#modsTable + 1] = { beatIndex + 6.40,	"Cross",	  modStrength*2,  0.2,	1 }
	modsTable[#modsTable + 1] = { beatIndex + 6.40,	"Reverse",	 -modStrength,    0.2,	1 } 
	modsTable[#modsTable + 1] = { beatIndex + 6.40,	"Cross",	 -modStrength*2,  0.2,  2 } 
	modsTable[#modsTable + 1] = { beatIndex + 6.40,	"Reverse",	  modStrength,    0.2,	2 } 
	modsTable[#modsTable + 1] = { beatIndex + 7.40,	"Cross",	 -modStrength*2,  0.2,  1 } 
	modsTable[#modsTable + 1] = { beatIndex + 7.40,	"Reverse",	  modStrength,    0.2,	1 } 
	modsTable[#modsTable + 1] = { beatIndex + 7.40,	"Cross",	  modStrength*2,  0.2,	2 } 
	modsTable[#modsTable + 1] = { beatIndex + 7.40,	"Reverse",	 -modStrength,    0.2,	2 } 
end


modsTable = SortModsTable(modsTable)
_FG_[#_FG_ + 1] = LoadActor("./modsHQ.lua", {modsTable, 0})


return _FG_
