-------------------------------------------------------------------------------
--
--		paraoka - "summerghost (long)"
--		Bubble Tea Invitational 2.5
--		
--		Author: 	Telperion
--		Date: 		2017-06-03
--		Target:		SM5.0.12+
--
-------------------------------------------------------------------------------
--
--		Gosh, it's warm!
--		The Japanese like to frighten each other with ghosts and ghouls in the
--		summer months, claiming that such spooks actually have a cooling effect
--		on the human body.
--		Well, I don't know how much I believe that, but I wish I had a chilling
--		experience to cut the heat a bit.
--
-------------------------------------------------------------------------------

local circumvention = false
if circumvention then
	return Def.ActorFrame {}
end

local niceSpeed = (420) / 141			-- This song is 141 BPM.
local expnSpeed = niceSpeed * 0.6
local slowSpeed = niceSpeed * 0.7
local sw = SCREEN_WIDTH
local sh = SCREEN_HEIGHT
local BPS = GAMESTATE:GetSongBPS()
local ofs = 0.009
local overtime = 0
local visualOffset = -0.01
local fgmsg = 0
local checked = false
local plr = {nil, nil}

local PI = math.pi
local LOG2 = math.log(2.0)
local DEG_TO_RAD = math.pi / 180.0



--
--	Texture constants
--
local ttw = math.pow(2, math.ceil(math.log(sw) / LOG2))
local tth = math.pow(2, math.ceil(math.log(sh) / LOG2))
local twscale = sw / ttw
local thscale = sh / tth

--
-- 	some funktion !
--
function SideSign(i) return (i == 2) and 1 or -1 end

function RangeScale(t, inLower, inUpper, outLower, outUpper)
	local ti = (t - inLower) / (inUpper - inLower)
	return outLower + ti * (outUpper - outLower)
end

function RangeClamp(t, outLower, outUpper)
	local ti = t
	ti = (ti < outLower) and outLower or ti
	ti = (ti > outUpper) and outUpper or ti
	return ti
end

function Whomst(act)
	return tonumber(string.match(act:GetName(), "P([0-9]+)"))
end

function Whichst(act)
	return tonumber(string.match(act:GetName(), "_([0-9]+)"))
end


function SortModsTable(mt)
	-- Insertion sort into the butt
	-- because wow I'm mostly lazy
	-- and can assume it's mostly sorted

	mtNew = {}

	for _,mr in pairs(mt) do
		local insertLocation = 1
		for i = #mtNew,1,-1 do
--			Trace("### SortModsTable: "..mtNew[i][1].." vs. "..mr[1])
			if mtNew[i][1] <= mr[1] then
				insertLocation = i+1
				break
			end
		end
--		Trace("### SortModsTable: "..mr[1].." ("..mr[2]..") -> "..insertLocation)
		table.insert(mtNew, insertLocation, mr)
	end

	return mtNew
end


local _DZ_ = Def.ActorFrame {}
local _FG_ = Def.ActorFrame {
	InitCommand = function(self)
	end,
	OnCommand = function(self)
		self:fov(45)
			:vanishpoint(sw/2, sh/4)			
			:SetDrawByZPosition(true)

		plr[1] = SCREENMAN:GetTopScreen():GetChild('PlayerP1')
		plr[2] = SCREENMAN:GetTopScreen():GetChild('PlayerP2')
		SCREENMAN:GetTopScreen():SetDrawByZPosition(true)


		local hamburger = SCREENMAN:GetTopScreen()
		if hamburger:GetScreenType() == "ScreenType_Gameplay" then
			hamburger:GetChild("Overlay" ):decelerate(4.0 / BPS):diffusealpha(0.0)
			hamburger:GetChild("Underlay"):decelerate(4.0 / BPS):diffusealpha(0.0)
		end
				
		local playersFound = 0
		for pn = 1,2 do
			pv = plr[pn]
			if pv then
				pv:GetChild("Combo"):hibernate(1573)
			end
			
			pops = GAMESTATE:GetPlayerState("PlayerNumber_P"..pn):GetPlayerOptions("ModsLevel_Song")
			pops:FailSetting('FailType_Off')
		end

		self:sleep(1573)
	end
};



-------------------------------------------------------------------------------
--
-- 		Summerghosting AMVs
--
local tcx = twscale * 0.5
local tcy = thscale * 0.5
local circleTicks = 40
local circleAngleTick = 2*PI / circleTicks

function CalculateCircleBaseVertices(innerRadius, outerRadius)
	verts = {}
	
	--
	-- 1--3--5--7-...
	-- |  |  |  |
	-- 2--4--6--8-...
	--
	local ttInnerRadiusX = innerRadius / ttw
	local ttInnerRadiusY = innerRadius / tth
	local ttOuterRadiusX = outerRadius / ttw
	local ttOuterRadiusY = outerRadius / tth
	for tateIndex = 0, circleTicks do
		local thetaL = tateIndex *  circleAngleTick
		local cosL = math.cos(thetaL)
		local sinL = math.sin(thetaL)

		verts[#verts+1] = {
			{cosL * innerRadius, 			sinL * innerRadius, 			0},
			Color.White,
			{cosL * ttInnerRadiusX + tcx, 	sinL * ttInnerRadiusY + tcy}
		}
		verts[#verts+1] = {
			{cosL * outerRadius, 			sinL * outerRadius, 			0},
			Color.White,
			{cosL * ttOuterRadiusX + tcx, 	sinL * ttOuterRadiusY + tcy}
		}
	end
	
	return verts
end


-- Instantiate!
circleRadii = {0, 0.045*sh, 0.135*sh, 0.225*sh, 0.315*sh, 0.405*sh, 0.495*sh}
circleActors = {{}, {}}

for i = 1,3 do
	_FG_[#_FG_ + 1] = Def.ActorMultiVertex {
		Name = "CircleLite"..i,
		InitCommand = function(self)
			circleActors[1][i] = self

			local verts = CalculateCircleBaseVertices(
					circleRadii[2*i],
					circleRadii[2*i + 1]
				)

			self:visible(false)
				:xy(sw/2, sh/2)
				:z(3)
				:SetVertices(verts)
				:SetDrawState{First = 1,
							  Num = (circleTicks + 1) * 2,
							  Mode = "DrawMode_QuadStrip"}
		end,
		SetVisibleCommand = function(self)
			self:visible(true)
				:diffusealpha(1.0)
		end,
	}
end

for i = 1,3 do
	_FG_[#_FG_ + 1] = Def.ActorMultiVertex {
		Name = "CircleDark"..i,
		InitCommand = function(self)
			circleActors[2][i] = self

			local verts = CalculateCircleBaseVertices(
					circleRadii[2*i - 1],
					circleRadii[2*i]
				)

			self:visible(false)
				:xy(sw/2, sh/2)
				:z(3)
				:SetVertices(verts)
				:SetDrawState{First = 1,
							  Num = (circleTicks + 1) * 2,
							  Mode = "DrawMode_QuadStrip"}
		end,
	}
end


--
-- 		Summerghosting AMVs
--
-------------------------------------------------------------------------------



-------------------------------------------------------------------------------
--
-- 		Proxies (as usual)
--

--
--	Player proxies (one apiece)
--
for pn = 1,2 do
	i = 1
	_FG_[#_FG_ + 1] = Def.ActorFrame {	
		Name = "ProxyP"..pn.."Outer",
		Def.ActorFrame {	
			Name = "ProxyP"..pn.."Inner",
			Def.ActorProxy {					
				Name = "ProxyP"..pn,
				InitCommand = function(self)
					self:aux( Whomst(self) )
				end,
				BeginCommand=function(self)
					local McCoy = SCREENMAN:GetTopScreen():GetChild('PlayerP'..self:getaux())
					if McCoy then self:SetTarget(McCoy) else self:hibernate(1573) end
				end,
				OnCommand=function(self)
					local McCoy = SCREENMAN:GetTopScreen():GetChild('PlayerP'..self:getaux())
					if McCoy then 
						self:xy(-McCoy:GetX(), -McCoy:GetY())
						self:GetParent():xy(McCoy:GetX(), McCoy:GetY())
					end
				end,
				RecenterProxyMessageCommand=function(self)					
					local McCoy = SCREENMAN:GetTopScreen():GetChild('PlayerP'..self:getaux())
					if McCoy then 
						self:xy(-McCoy:GetX(), -McCoy:GetY())
						self:GetParent():xy(McCoy:GetX(), McCoy:GetY())
					end
				end
			},
			InitCommand = function(self)
--				self:aux( Whichst(self) )
			end,
			OnCommand = function(self)
			end,
		},
		InitCommand = function(self)
--			self:aux( Whichst(self) )
			self:visible(false)
		end,
		OnCommand = function(self)
			self:xy(0, 0)
				:z(1)
		end,
	}
end



--
--	Texture proxies (one apiece)
--
playerTextureActors = {}
playerBGActors = {}
playerBGModels = {nil, nil}

for pn = 1,2 do
	i = 1
	_FG_[#_FG_ + 1] = Def.ActorFrameTexture {	
		Name = "TextureProxyP"..pn.."Outer",
		Def.ActorFrame {	
			Name = "TextureProxyP"..pn.."Inner",
			Def.ActorProxy {					
				Name = "TextureProxyP"..pn,
				InitCommand = function(self)
					self:aux( Whomst(self) )
				end,
				BeginCommand=function(self)
					local McCoy = SCREENMAN:GetTopScreen():GetChild('PlayerP'..self:getaux())
					if McCoy then self:SetTarget(McCoy) else self:hibernate(1573) end
				end,
				OnCommand=function(self)
					local McCoy = SCREENMAN:GetTopScreen():GetChild('PlayerP'..self:getaux())
					if McCoy then 
						self:xy(-McCoy:GetX(), -McCoy:GetY())
						self:GetParent():xy(McCoy:GetX(), McCoy:GetY())
					end
				end,
				RecenterProxyMessageCommand=function(self)					
					local McCoy = SCREENMAN:GetTopScreen():GetChild('PlayerP'..self:getaux())
					if McCoy then 
						self:xy(-McCoy:GetX(), -McCoy:GetY())
						self:GetParent():xy(McCoy:GetX(), McCoy:GetY())
					end
				end
			},
			InitCommand = function(self)
--				self:aux( Whichst(self) )
			end,
			OnCommand = function(self)
			end,
		},
		InitCommand = function(self)
--			self:aux( Whichst(self) )
			self:SetTextureName(self:GetName())
				:SetWidth(sw)
				:SetHeight(sh)
				:EnableAlphaBuffer(true)
				:Create()
		end,
		BeginCommand = function(self)
			if playerTextureActors[Whomst(self)] then
				playerTextureActors[Whomst(self)]:SetTexture(self:GetTexture())
			end
		end,
		OnCommand = function(self)
			self:xy(0, 0)
				:z(1)
		end,
	}

	_FG_[#_FG_ + 1] = Def.ActorFrameTexture {
		Name = "ProxyAndBG_P"..pn,
	
		Def.Sprite {
			Name = "UnderlayBG_P"..pn,
			Texture = "BTI2p5-bg-dark.png",
			InitCommand = function(self)
				self:xy(sw / 2, sh / 2)
					:SetWidth(sw)
					:SetHeight(sw * 0.75)
					:z(0)
					:diffuse(0.3, 0.3, 0.3, 0.0)
		--			:zbuffer(1)
		--			:ztestmode("ZTestMode_WriteOnFail")
			end,

			EngageTripletingMessageCommand = function(self, args)
				local draggedPlayer = args[1] and args[1] or 2
				local engageTime 	= args[2] and args[2] or 0

--				self:SetTexture( playerBGModels[Whichst(self)]:GetTexture() )
				self:decelerate(engageTime / BPS)
					:diffusealpha(1.0)
			end,

			DisengageTripletingMessageCommand = function(self, args)
				local disengageTime = args[1] and args[1] or 0

				self:accelerate(disengageTime / BPS)
					:diffusealpha(0.0)
			end,

		},

		Def.Sprite {
			Name = "SpriteProxyP"..pn,
			InitCommand = function(self)
				playerTextureActors[Whomst(self)] = self

				self:xy(sw / 2, sh / 2)
					:SetWidth(sw)
					:SetHeight(sh)
					:z(1)

				if Whomst(self) == 1 then
--					self:diffuse(0.0, 0.0, 0.0, 0.8)
						--:blend("BlendMode_Subtract")
				else
--					self:glow(1.0, 1.0, 1.0, 1.0)
						--:blend("BlendMode_Add")
				end
			end,
			DarkenProxyMessageCommand = function(self, args)
				local pn 	   	= (#args >= 1) and args[1] or 1	-- player to become the echo
				local startBeat	= (#args >= 2) and args[2] or 0	-- start of model rhythm
				local duration 	= (#args >= 3) and args[3] or 8	-- duration of echoed portion

				if pn == Whomst(self) then
					self:diffuse(0.0, 0.0, 0.0, 0.8)
						:sleep(2 * duration / BPS)
						:queuecommand("ResetShade")

					if playerBGActors[pn] then
						playerBGActors[pn]:addz(-1)
					end
				end
			end,
			WhitenProxyMessageCommand = function(self, args)
				local pn 	   	= (#args >= 1) and args[1] or 1	-- player to become the echo
				local startBeat	= (#args >= 2) and args[2] or 0	-- start of model rhythm
				local duration 	= (#args >= 3) and args[3] or 8	-- duration of echoed portion

				if pn == Whomst(self) then
					self:glow(1.0, 1.0, 1.0, 1.0)
						:sleep(2 * duration / BPS)
						:queuecommand("ResetShade")

					if playerBGActors[pn] then
						playerBGActors[pn]:addz(-1)
					end
				end
			end,
			ResetShadeCommand = function(self)
				self:glow(1.0, 1.0, 1.0, 0.0)
					:diffuse(1.0, 1.0, 1.0, 1.0)


				if playerBGActors[Whomst(self)] then
					playerBGActors[Whomst(self)]:addz(1)
				end
			end,
		},

		InitCommand = function(self)
			self:SetTextureName(self:GetName())
				:SetWidth(sw)
				:SetHeight(sh)
				:EnableAlphaBuffer(true)
				:Create()
		end,
		BeginCommand = function(self)
			self:SetDrawByZPosition(true)

			if Whomst(self) == 1 then
				for circleIndex = 1,3 do
					if circleActors[1][circleIndex] then
						circleActors[1][circleIndex]:SetTexture( self:GetTexture() )
					end
				end
			else
				for circleIndex = 1,3 do
					if circleActors[2][circleIndex] then
						circleActors[2][circleIndex]:SetTexture( self:GetTexture() )
					end
				end
			end
			
			playerBGActors[Whomst(self)]:SetTexture( self:GetTexture() )
							 			:customtexturerect(0, 0, twscale, thscale)
		end,
	}
end


--
--	Judgment proxies
--
for pn = 1,2 do
	_FG_[#_FG_ + 1] = Def.ActorProxy {
		Name = "JudgeP"..pn.."Proxy",
		InitCommand = function(self)
			self:aux( Whomst(self) )
		end,
		BeginCommand = function(self)
			local McCoy = SCREENMAN:GetTopScreen():GetChild('PlayerP'..self:getaux())
			if McCoy then 
				local McJudge = McCoy:GetChild('Judgment')
				self:SetTarget(McJudge)
				McJudge:visible(false)
			else 
				self:hibernate(1573)
			end
		end,
		OnCommand = function(self)
			local p = self:GetParent()
						  :GetChild('ProxyP'..self:getaux().."Outer")
						  :GetChild('ProxyP'..self:getaux().."Inner")
			
			self:xy(p:GetX(), sh/2)
				:z(5)
		end,
	}
end

--
-- 		Proxies (as usual)
--
-------------------------------------------------------------------------------


-------------------------------------------------------------------------------
--
--		Extra nuts 'n' bolts
--

local texArrows = {
	0.000,		-- 4th
	0.375,		-- 16th
	0.125,		-- 8th
	0.250,		-- 12th
	0.625,		-- 32nd
	0.875,		-- 64th
--	0.500,		-- 24th
--	0.750,		-- 48th
	}			-- Texture coordinate shifts

local quantIndex = 3
_DZ_[#_DZ_ + 1] = NOTESKIN:LoadActorForNoteSkin("Down", "Tap Note", "cyber") .. {
	Name = "SpectrumNote"..quantIndex,
	InitCommand = function(self)
		self:visible(true)
			:xy(sw / 2, sh / 2)
--	 		:zbuffer(1)
--			:zbias(1)
--			:zwrite(true)
			:texturetranslate(texArrows[quantIndex], 0)
	end,
	TestMaskVolatileMessageCommand = function(self, args)
		if #args < 1 then
			args[1] = 16.0
		end
		if #args < 2 then
			args[2] = 360
		end

		self:decelerate(args[1] / BPS)
			:rotationz(args[2])
	end,
}



local ReceptorActors = {}
local ReceptorHolders = {}
local ReceptorRings = {}
local ReceptorRingTextures = {}
local nReceptors = 6
local nReceptorRings = 6


for rri = 1,nReceptorRings do
	local ReceptorRingTexCnstr = Def.ActorFrameTexture {
		Name = "ReceptorRingTexture_"..rri,
		InitCommand = function(self)
			ReceptorRingTextures[rri] = self

			self:SetTextureName(self:GetName())
				:SetWidth(sw)
				:SetHeight(sh)
				:EnableAlphaBuffer(true)
				:Create()
		end,
		BeginCommand = function(self)
			if ReceptorRings[Whichst(self)] then
				ReceptorRings[Whichst(self)]:SetTexture(self:GetTexture())
			end
		end,
	}

	ReceptorActors[rri] = {}
	ReceptorHolders[rri] = {}
	for rcpi = 1,nReceptors do
		ReceptorRingTexCnstr[#ReceptorRingTexCnstr + 1] = Def.ActorFrame {
			Name = "RCP_Holder_"..rcpi,
			InitCommand = function(self)
				ReceptorHolders[rri][rcpi] = self
				local mi = Whichst(self)
				self:xy(sw / 2, sh / 2)
					:rotationz((mi + 0.5) * 360 / nReceptors)
			end,
			PulseCommand = function(self)
				self:linear(4 / BPS)
					:zoom(0.0)
					:queuecommand("Reset")
			end,
			ResetCommand = function(self)
				self:finishtweening()
					:zoom(1.0)
			end,
			NOTESKIN:LoadActorForNoteSkin("Down", "Receptor", "cyber") .. {
				Name = "RCP_"..rcpi,
				InitCommand = function(self)
					ReceptorActors[rri][rcpi] = self
					local mi = Whichst(self)
					self:xy(0, 0)
				end,
				PulseCommand = function(self)
					self:decelerate(4 / BPS)
						:y(128)
						:queuecommand("Reset")
				end,
				ResetCommand = function(self)
					self:finishtweening()
						:y(0)
				end,
			}
		}
	end

	_FG_[#_FG_ + 1] = ReceptorRingTexCnstr
	_FG_[#_FG_ + 1] = Def.Sprite {
		Name = "ReceptorRing_"..rri,
		InitCommand = function(self)
			ReceptorRings[rri] = self
			self:visible(false)
		end,
		FlashyPulseReceptorMessageCommand = function(self, args)
			-- [1]: which one to pulse and move
			-- [2]: x-coordinate
			-- [3]: y-coordinate

			if args[1] == Whichst(self) then
				self:rotationz(math.random() * 360)
				self:xy(args[2], args[3])
				if #args >= 4 then
					self:queuecommand("PulseAway")
				else
					self:queuecommand("Pulse")
				end
				ReceptorRingTextures[args[1]]:queuecommand("Pulse")
				for rcpi = 1,nReceptors do					
					ReceptorHolders[args[1]][rcpi]:queuecommand("Pulse")
					ReceptorActors[args[1]][rcpi]:queuecommand("Pulse")
				end
			end
		end,
		PulseCommand = function(self)
			self:visible(true)
				:diffusealpha(0.0)
				:decelerate(0.5 / BPS)
				:diffusealpha(1.0)
				:queuecommand("Pulse2")
		end,
		Pulse2Command = function(self)
			self:accelerate(2.5 / BPS)
				:diffusealpha(0.0)
				:queuecommand("Reset")
		end,
		PulseAwayCommand = function(self)
			self:visible(true)
				:diffusealpha(0.0)
				:decelerate(0.5 / BPS)
				:diffusealpha(1.0)
				:queuecommand("PulseAway2")
		end,
		PulseAway2Command = function(self)
			self:accelerate(1.5 / BPS)
				:y(sh * 1.5)
				:diffusealpha(0.0)
				:queuecommand("Reset")
		end,
		ResetCommand = function(self)
			self:finishtweening()
				:visible(false)
		end,
	}
end



_FG_[#_FG_ + 1] = Def.Sprite {
	Name = "MaskSpriteP1",
	Texture = "BTI2p5-bg-lite.png",
	InitCommand = function(self)
		playerBGActors[1] = self

		self:xy(sw / 2, sh / 2)
			:SetWidth(sw)
			:SetHeight(sw * 0.75)
			:z(2.01)
			:diffuse(1.0, 1.0, 1.0, 1.0)
	end,

}

_FG_[#_FG_ + 1] = Def.Sprite {
	Name = "MaskSpriteP2",
	Texture = "BTI2p5-bg-dark.png",
	InitCommand = function(self)
		playerBGActors[2] = self

		self:xy(sw / 2, sh / 2)
			:SetWidth(sw)
			:SetHeight(sw * 0.75)
			:z(2)
			:diffuse(1.0, 1.0, 1.0, 1.0)
	end,

}

_FG_[#_FG_ + 1] = Def.Sprite {
	Name = "JustBackground",
	Texture = "BTI2p5-bg-lite.png",
	InitCommand = function(self)
		playerBGModels[1] = self
		self:xy(sw / 2, sh / 2)
			:SetWidth(sw)
			:SetHeight(sw * 0.75)
			:z(-2)
			:diffuse(0.3, 0.3, 0.3, 1.0)
	end,

}

_FG_[#_FG_ + 1] = Def.Sprite {
	Name = "JustAnotherBackground",
	Texture = "BTI2p5-bg-dark.png",
	InitCommand = function(self)
		playerBGModels[2] = self
		self:xy(sw / 2, sh / 2)
			:SetWidth(sw)
			:SetHeight(sw * 0.75)
			:z(-2.01)
			:diffuse(0.3, 0.3, 0.3, 1.0)
	end,

}

local weewooSpinner = 0
local weewooOrigin = {{sw/2, sh/2}, {sw/2, sh/2}}
local weewooDisplacement = sh/3
local weewooRotationMax = 45
local weewooRotationCount = 8
local SpinControllerAuxActor = nil


_FG_[#_FG_ + 1] = Def.Actor {
	InitCommand = function(self)
		SpinControllerAuxActor = self
		self:aux(0)
	end,


	WeeWooActivateMessageCommand = function(self, args)
		local spinPN   	= (#args >= 1) and args[1] or 1	-- player to do the spinny bit
		local duration 	= (#args >= 2) and args[2] or 8	-- duration of echoed portion

		weewooSpinner = spinPN
		for pn = 1,2 do
			weewooOrigin[pn][1] = playerBGActors[pn]:GetX()
			weewooOrigin[pn][2] = playerBGActors[pn]:GetY()
		end

		self:finishtweening()
			:accelerate(duration / BPS)
			:aux(1)
	end,

	WeeWooDeactivateMessageCommand = function(self, args)	
		weewooSpinner = 0
		for pn = 1,2 do
			playerBGActors[pn]:xy(weewooOrigin[pn][1], weewooOrigin[pn][2])
							  :rotationz(0)
							  :rotationy(0)
							  :diffusealpha(1)
		end

		self:finishtweening()
			:aux(0)
	end,
}



-------------------------------------------------------------------------------
--
-- 		This is where the shit will be happening.
--

local moveToCenter_Duration  = 4
local moveToCenter_Origins   = {0, 0}
local moveToCenter_UnitDist  = {0, 0}
local moveToCenter_Quants	 = {0.000, 0.750, 1.500, 
								2.000, 2.750, 
								3.000, 3.125, 3.250, 3.375, 3.500, 3.625, 3.750, 3.875, 4.000}

local messageList = {
	{  4.00, "RecenterProxy"},
	{  4.00, "RotateWholeField",	{1, 0, 1}},
	{  4.00, "RotateWholeField",	{2, 0, 1}},
--	{ 16.00, "TestMaskVolatile",	{32, 720}},

	{  4.000, "FlashyPulseReceptor",	{1, sw*0.2, sh*0.1}},
	{  4.000, "FlashyPulseReceptor",	{2, sw*0.8, sh*0.1}},
	{  6.000, "FlashyPulseReceptor",	{3, sw*0.1, sh*0.9}},
	{  6.000, "FlashyPulseReceptor",	{4, sw*0.9, sh*0.9}},
	{  8.000, "FlashyPulseReceptor",	{5, sw*0.3, sh*0.8}},
	{  8.000, "FlashyPulseReceptor",	{6, sw*0.7, sh*0.8}},
	{ 10.000, "FlashyPulseReceptor",	{1, sw*0.1, sh*0.2}},
	{ 10.000, "FlashyPulseReceptor",	{2, sw*0.9, sh*0.2}},
	{ 12.000, "FlashyPulseReceptor",	{3, sw*0.2, sh*0.5}},
	{ 12.000, "FlashyPulseReceptor",	{4, sw*0.8, sh*0.5}},
	{ 14.000, "FlashyPulseReceptor",	{5, sw*0.1, sh*0.8}},
	{ 14.000, "FlashyPulseReceptor",	{6, sw*0.9, sh*0.8}},
	{ 16.000, "FlashyPulseReceptor",	{1, sw*0.3, sh*0.6}},
	{ 16.000, "FlashyPulseReceptor",	{2, sw*0.7, sh*0.6}},

	{ 18.000, "FlashyPulseReceptor",	{3, sw*0.25,  80}},	-- Actual receptors appear!
	{ 18.000, "FlashyPulseReceptor",	{4, sw*0.75,  80}},	-- Actual receptors appear!

	{ 20.000, "FlashyPulseReceptor",	{5, sw*0.25-96, 320}},
	{ 20.000, "FlashyPulseReceptor",	{6, sw*0.75-96, 320}},
	{ 22.000, "FlashyPulseReceptor",	{1, sw*0.25-32, 320}},
	{ 22.000, "FlashyPulseReceptor",	{2, sw*0.75-32, 320}},
	{ 24.000, "FlashyPulseReceptor",	{4, sw*0.25+32, 320}},
	{ 24.000, "FlashyPulseReceptor",	{3, sw*0.75+32, 320}},
	{ 26.000, "FlashyPulseReceptor",	{5, sw*0.25+96, 320}},
	{ 26.000, "FlashyPulseReceptor",	{6, sw*0.75+96, 320}},
	{ 28.000, "FlashyPulseReceptor",	{1, sw*0.5,  80}},
	{ 28.000, "FlashyPulseReceptor",	{2, sw*0.5, 400}},
	{ 30.000, "FlashyPulseReceptor",	{3, sw*0.5, 240, 1}},



	{ 31.000, "MoveToCenterStage"},	
	{ 32.000, "MoveToCenterStep",	{ 0.000}},
	{ 32.750, "MoveToCenterStep",	{ 0.750}},
	{ 33.500, "MoveToCenterStep",	{ 1.500}},
	{ 34.000, "MoveToCenterStep",	{ 2.000}},
	{ 34.750, "MoveToCenterStep",	{ 2.750}},
	{ 35.000, "MoveToCenterStep",	{ 3.000}},
	{ 35.125, "MoveToCenterStep",	{ 3.125}},
	{ 35.250, "MoveToCenterStep",	{ 3.250}},
	{ 35.375, "MoveToCenterStep",	{ 3.375}},
	{ 35.500, "MoveToCenterStep",	{ 3.500}},
	{ 35.625, "MoveToCenterStep",	{ 3.625}},
	{ 35.750, "MoveToCenterStep",	{ 3.750}},
	{ 35.875, "MoveToCenterStep",	{ 3.875}},
	{ 36.000, "MoveToCenterStep",	{ 4.000}},

	{ 40.000, "RecenterProxy"},


	{421.000, "WeeWooActivate",		{1, 7}},
	{428.000, "WeeWooDeactivate",	{}},
	{429.000, "WeeWooActivate",		{2, 7}},
	{436.000, "WeeWooDeactivate",	{}},
	{437.000, "WeeWooActivate",		{1, 7}},
	{444.000, "WeeWooDeactivate",	{}},
	{445.000, "WeeWooActivate",		{2, 7}},
	{452.000, "WeeWooDeactivate",	{}},

	{ 35.750, "EchoRhythm",			{2, 0, 8, 0}},
	{ 55.750, "EchoRhythm",			{2, 0, 4, 0}},
	{ 67.750, "EchoRhythm",			{2, 0, 8, 1}},
	{ 87.750, "EchoRhythm",			{2, 0, 4, 1}},
	{101.250, "EchoRhythm",			{1, 0, 8, 0}},
	{121.250, "EchoRhythm",			{1, 0, 4, 0}},
	{131.750, "EchoRhythm",			{1, 0, 8, 1}},
	{151.750, "EchoRhythm",			{1, 0, 4, 1}},

	{260.000, "EchoRhythm",			{2, 0, 8, 1}},
	{280.000, "EchoRhythm",			{1, 0, 4, 0}},

--	{378.000, "EchoRhythm",			{2, 0, 2, 0}},
--	{402.000, "EchoRhythm",			{1, 0, 2, 1}},

	{259.000, "RotateWholeField",	{1, 0, 1}},
	{259.000, "RotateWholeField",	{2, 0, 1}},

	{507.000, "EngageTripleting",		{2, 1}},
	{508.000, "PerformTripleting",		{2, 32, 22.5}},
	{508.000, "DarkenProxy",			{2, 0, 16}},		-- THIS IS NOT HOW I WAS SUPPOSED TO USE IT
	{540.000, "DisengageTripleting",	{   1}},


--	{ 27.000, "EngageTripleting",		{2, 1}},
--	{ 28.000, "PerformTripleting",		{2, 16, 45}},
--	{ 40.000, "DisengageTripleting",	{   1}},
};

for beatIndex = 36,84,16 do	
	messageList[#messageList + 1] = { beatIndex + 0.00, "FlashyPulseReceptor",	{1, sw*0.3, sh*0.3}}
	messageList[#messageList + 1] = { beatIndex + 0.00, "FlashyPulseReceptor",	{2, sw*0.7, sh*0.7}}
	messageList[#messageList + 1] = { beatIndex + 0.00, "FlashyPulseReceptor",	{3, sw*0.2, sh*0.8}}
	messageList[#messageList + 1] = { beatIndex + 0.00, "FlashyPulseReceptor",	{4, sw*0.8, sh*0.2}}

	messageList[#messageList + 1] = { beatIndex + 8.00, "FlashyPulseReceptor",	{1, sw*0.3, sh*0.7}}
	messageList[#messageList + 1] = { beatIndex + 8.00, "FlashyPulseReceptor",	{2, sw*0.7, sh*0.3}}
	messageList[#messageList + 1] = { beatIndex + 8.00, "FlashyPulseReceptor",	{3, sw*0.2, sh*0.2}}
	messageList[#messageList + 1] = { beatIndex + 8.00, "FlashyPulseReceptor",	{4, sw*0.8, sh*0.8}}
end

local rwfSeverity = 0.2
for beatIndex = 180, 244, 16 do
	messageList[#messageList + 1] = { beatIndex +  0.00,	"RotateWholeField",	{1, rwfSeverity * -1.5 * PI/6, 1}}
	messageList[#messageList + 1] = { beatIndex +  0.00,	"RotateWholeField",	{2, rwfSeverity * -1.5 * PI/6, 1}}
	messageList[#messageList + 1] = { beatIndex +  2.50,	"RotateWholeField",	{1, rwfSeverity *  1.5 * PI/6, 1}}
	messageList[#messageList + 1] = { beatIndex +  2.50,	"RotateWholeField",	{2, rwfSeverity *  1.5 * PI/6, 1}}
	messageList[#messageList + 1] = { beatIndex +  4.00,	"RotateWholeField",	{1, rwfSeverity * -0.5 * PI/6, 1}}
	messageList[#messageList + 1] = { beatIndex +  4.00,	"RotateWholeField",	{2, rwfSeverity * -0.5 * PI/6, 1}}
	messageList[#messageList + 1] = { beatIndex +  6.50,	"RotateWholeField",	{1, rwfSeverity *  0.5 * PI/6, 1}}
	messageList[#messageList + 1] = { beatIndex +  6.50,	"RotateWholeField",	{2, rwfSeverity *  0.5 * PI/6, 1}}
	messageList[#messageList + 1] = { beatIndex +  8.00,	"RotateWholeField",	{1, rwfSeverity * -1.5 * PI/6, 1}}
	messageList[#messageList + 1] = { beatIndex +  8.00,	"RotateWholeField",	{2, rwfSeverity * -1.5 * PI/6, 1}}
	messageList[#messageList + 1] = { beatIndex + 10.50,	"RotateWholeField",	{1, rwfSeverity *  1.5 * PI/6, 1}}
	messageList[#messageList + 1] = { beatIndex + 10.50,	"RotateWholeField",	{2, rwfSeverity *  1.5 * PI/6, 1}}
	messageList[#messageList + 1] = { beatIndex + 12.00,	"RotateWholeField",	{1, rwfSeverity * -1.0 * PI/6, 1}}
	messageList[#messageList + 1] = { beatIndex + 12.00,	"RotateWholeField",	{2, rwfSeverity * -1.0 * PI/6, 1}}
	messageList[#messageList + 1] = { beatIndex + 14.50,	"RotateWholeField",	{1, rwfSeverity *  1.0 * PI/6, 1}}
	messageList[#messageList + 1] = { beatIndex + 14.50,	"RotateWholeField",	{2, rwfSeverity *  1.0 * PI/6, 1}}

	rwfSeverity = rwfSeverity + 0.2
end


messageList = SortModsTable(messageList)

gfxUpdateFunction = function()
	-- Most things are determined by beat, believe it or not.		
	overtime = GAMESTATE:GetSongBeat() - visualOffset
	
	-- TODO: this assumes the effect applies over a constant BPM section!!
	BPS = GAMESTATE:GetSongBPS()
	
	-- Initializations0
	if overtime >=   0.0 and not checked then
		for i,v in ipairs(plr) do
			if v then
				v:visible(false):decelerate(4.0 / BPS)
--				 :x(sw/2)
				 :y(sh/2 - 30)
--				 :y(sh/2 - SideSign(i) * 60)
				 :z(-i)
			end
		end

		checked = true
	end
			
	-- Broadcast messages on their own terms.
	while true do
		if fgmsg < #messageList then
			messageBeat, messageName, messageArgs = unpack(messageList[fgmsg+1])
			if overtime >= messageBeat then			
				if messageArgs then
					MESSAGEMAN:Broadcast( messageName, messageArgs )
				else
					MESSAGEMAN:Broadcast( messageName )
				end
				
				fgmsg = fgmsg + 1
			else
				break
			end
		else
			break
		end
	end

--	for circleIndex = 1,3 do
--		if circleActors[1][circleIndex] then
--			circleActors[1][circleIndex]:rotationz(45 * overtime)
--		end
--	end


	-- weewoo feeling
	if weewooSpinner ~= 0 then
		local thisTime = SpinControllerAuxActor:getaux()
		local pnI =     weewooSpinner
		local pnU = 3 - weewooSpinner
		playerBGActors[pnI]:rotationz(math.sin(thisTime * 2*PI * weewooRotationCount) * weewooRotationMax * thisTime)
						   :rotationy(math.cos(thisTime * 2*PI * weewooRotationCount) * weewooRotationMax * thisTime)
						   :y(weewooOrigin[pnI][2] - weewooDisplacement * thisTime)
						   :diffusealpha(1 - thisTime)
		playerBGActors[pnU]:y(weewooOrigin[pnU][2] + weewooDisplacement * (1 - thisTime))
						   :diffusealpha(    thisTime)

	end


--	for pn = 1,2 do
--		for circleIndex = 1,3 do
--			if circleActors[pn][circleIndex] then
--				circleActors[pn][circleIndex]:rotationz(30 * SideSign(pn) * math.sin(overtime * PI * 0.5))
--			end
--		end
--	end
--	Trace('Update @ ' .. GAMESTATE:GetSongBeat() .. " (" .. GAMESTATE:GetCurMusicSeconds() .. " sec.)")
end


-------------------------------------------------------------------------------
--
-- Messaging controller
--

_FG_[#_FG_ + 1] = Def.ActorFrame {
	DiscoTimeMessageCommand = function(self, args)
		discoStartTime = overtime
		if #args >= 1 then
			discoTweenTime = args[1]
		end
		if #args >= 2 then
			discoProxyPeriod = args[2]
		end
	end,


	MoveToCenterStageMessageCommand = function(self, args)
		for i,v in ipairs(plr) do
			if v then
				moveToCenter_Origins[i]  = v:GetX()
				moveToCenter_UnitDist[i] = (sw/2 - moveToCenter_Origins[i]) / moveToCenter_Duration
			end
		end
	end,

	MoveToCenterStepMessageCommand = function(self, args)
		for i,v in ipairs(plr) do
			if v then
				v:x(moveToCenter_Origins[i] + moveToCenter_UnitDist[i] * args[1])
			end
		end
	end,

	--
	-- add a mod line that switches stealth+dark off and on appropriately!
	--
	EchoRhythmMessageCommand = function(self, args)
		local pn 	   	= (#args >= 1) and args[1] or 1	-- player to become the echo
		local startBeat	= (#args >= 2) and args[2] or 0	-- start of model rhythm
		local duration 	= (#args >= 3) and args[3] or 8	-- duration of echoed portion
		local echoType 	= (#args >= 4) and args[4] or 0	-- 0 = black, 1 = white

		if echoType == 1 then
			MESSAGEMAN:Broadcast("WhitenProxy", {pn, startBeat, duration})
		else
			MESSAGEMAN:Broadcast("DarkenProxy", {pn, startBeat, duration})
		end
	end,


	-- Control the rotation splines
	RotateWholeFieldMessageCommand = function(self, args)
		local pn 			= (#args >= 1) and args[1] or 1
		local endingAngle 	= (#args >= 2) and args[2] or PI
		local tweenTime 	= (#args >= 3) and args[3] or 4

		local pp = SCREENMAN:GetTopScreen():GetChild('PlayerP'..pn)
		local nf = pp:GetChild('NoteField')

		for i,cv in ipairs(nf:GetColumnActors()) do
			cv:smooth(tweenTime / BPS)
			cv:GetRotHandler():SetSplineMode('NoteColumnSplineMode_Offset')
			cv:GetRotHandler():SetBeatsPerT(1.3 - i * 0.2)
			local splr = cv:GetRotHandler():GetSpline()
			splr:SetSize(2)
			splr:SetPoint(1, {0, 0, -endingAngle})
			splr:SetPoint(2, {0, 0, -endingAngle})
			splr:Solve()
		end

		if playerBGActors[pn] then
			playerBGActors[pn]:decelerate(tweenTime/BPS)
							  :rotationz(endingAngle / DEG_TO_RAD)
		end
	end,

	-- Add dark + flip commands to correspond for each player!!
	EngageTripletingMessageCommand = function(self, args)
		local draggedPlayer = args[1] and args[1] or 2
		local engageTime 	= args[2] and args[2] or 0

		for circleIndex = 1,3 do							
			if circleActors[3 - draggedPlayer][circleIndex] then
				circleActors[3 - draggedPlayer][circleIndex]
					:visible(true)
					:diffusealpha(1.0)
			end

			if circleActors[draggedPlayer][circleIndex] then
				circleActors[draggedPlayer][circleIndex]
					:visible(true)
					:diffusealpha(0.0)
--					:baserotationz(180)
					:accelerate(engageTime / BPS)
					:diffusealpha(1.0)
			end
		end
	end,

	PerformTripletingMessageCommand = function(self, args)
		local draggedPlayer = args[1] and args[1] or 2
		local effectLength 	= args[2] and args[2] or 32
		local avgDegPerBeat	= args[3] and args[3] or 22.5
		local endingAngleDeg	= effectLength * avgDegPerBeat
		local endingAngleRad	= effectLength * avgDegPerBeat * DEG_TO_RAD

		for pn = 1,2 do
			for circleIndex = 1,3 do
				if circleActors[pn][circleIndex] then
					circleActors[pn][circleIndex]
						:linear(effectLength / BPS)
						:rotationz(endingAngleDeg * SideSign(pn))
				end
			end

			local pp = SCREENMAN:GetTopScreen():GetChild('PlayerP'..pn)
			local nf = pp:GetChild('NoteField')

			for i,cv in ipairs(nf:GetColumnActors()) do
				cv:linear(effectLength / BPS)
				cv:GetRotHandler():SetSplineMode('NoteColumnSplineMode_Offset')
				local splr = cv:GetRotHandler():GetSpline()
				splr:SetSize(2)
				splr:SetPoint(1, {0, 0, -endingAngleRad * SideSign(pn)})
				splr:SetPoint(2, {0, 0, -endingAngleRad * SideSign(pn)})
				splr:Solve()
			end
		end
	end,

	DisengageTripletingMessageCommand = function(self, args)
		local disengageTime 	= args[1] and args[1] or 0

		for pn = 1,2 do
			for circleIndex = 1,3 do
				if circleActors[pn][circleIndex] then
					circleActors[pn][circleIndex]
						:accelerate(disengageTime / BPS)
						:diffusealpha(0.0)
				end
			end
		end
	end,
}

local echoOffsetExt = 1.5
function EchoRhythmAddMods(mt, args)
	local pn 		= (#args >= 1) and args[1] or 1	-- player to become the echo
	local startBeat	= (#args >= 2) and args[2] or 0	-- start of model rhythm
	local duration 	= (#args >= 3) and args[3] or 8	-- duration of echoed portion

	local pnOther  = 3 - pn
	local offsetTimePerUnit = 2.5 / niceSpeed
	--	Arrows move at 64 pixels/beat at 1x
	--	One unit of offset spans 160 pixels
	local offsetTime = (echoOffsetExt + 1) * offsetTimePerUnit
	Trace (">>> EchoRhythmAddMods: pn = "..pn..", start = "..startBeat..", duration = "..duration..", offsetTime = "..offsetTime)

	--		|			|					|			|
	--  p 	stealth 	hidden, ofs hi->lo 	normal 		sudden, ofs hi->lo
	--	p'	normal		sudden, ofs hi->lo	stealth 	hidden, ofs hi->lo

	-- Prepare at beginning of model
	mt[#mt + 1] = {startBeat,		"Stealth",		1.0,			duration * 0.5,	pn}
	mt[#mt + 1] = {startBeat,		"Dark",			1.0,			duration * 0.5,	pn}
	mt[#mt + 1] = {startBeat,		"HiddenOffset",	echoOffsetExt,	0.0,			pn}
	mt[#mt + 1] = {startBeat,		"SuddenOffset",	echoOffsetExt,	0.0,			pnOther}

	-- Lead-in to echo
	mt[#mt + 1] = {startBeat 
					+ duration
					- offsetTime,	"Stealth",		0.0,			0.5,		pn}
	mt[#mt + 1] = {startBeat 
					+ duration
					- offsetTime,	"Dark",			0.0,			0.5,		pn}
	mt[#mt + 1] = {startBeat 
					+ duration
					- offsetTime,	"Hidden",		1.0,			0.0,		pn}
	mt[#mt + 1] = {startBeat 
					+ duration
					- offsetTime,	"HiddenOffset",-1.0,			offsetTime,	pn}

	mt[#mt + 1] = {startBeat 
					+ duration
					- offsetTime,	"Sudden",		1.0,			0.0,		pnOther}
	mt[#mt + 1] = {startBeat 
					+ duration
					- offsetTime,	"SuddenOffset",-1.0,			offsetTime,	pnOther}

	-- Prepare at beginning of echo
	mt[#mt + 1] = {startBeat
					+ duration,		"Stealth",		1.0,			0.0,		pnOther}
	mt[#mt + 1] = {startBeat
					+ duration,		"Dark",			1.0,			0.0,		pnOther}
	mt[#mt + 1] = {startBeat
					+ duration,		"Hidden",		0.0,			0.0,		pn}
	mt[#mt + 1] = {startBeat
					+ duration,		"Sudden",		0.0,			0.0,		pnOther}
	mt[#mt + 1] = {startBeat
					+ duration,		"HiddenOffset",	echoOffsetExt,	0.0,		pnOther}
	mt[#mt + 1] = {startBeat
					+ duration,		"SuddenOffset",	echoOffsetExt,	0.0,		pn}

	-- Trail-out of echo
	mt[#mt + 1] = {startBeat 
					+ 2*duration
					- offsetTime,	"Stealth",		0.0,			0.5,		pnOther}
	mt[#mt + 1] = {startBeat 
					+ 2*duration
					- offsetTime,	"Dark",			0.0,			0.5,		pnOther}
	mt[#mt + 1] = {startBeat 
					+ 2*duration
					- offsetTime,	"Hidden",		1.0,			0.0,		pnOther}
	mt[#mt + 1] = {startBeat 
					+ 2*duration
					- offsetTime,	"HiddenOffset",-1.0,			offsetTime,	pnOther}

	mt[#mt + 1] = {startBeat 
					+ 2*duration
					- offsetTime,	"Sudden",		1.0,			0.0,		pn}
	mt[#mt + 1] = {startBeat 
					+ 2*duration
					- offsetTime,	"SuddenOffset",-1.0,			offsetTime,	pn}

	-- OK, everything's done, kill off the persistent mods
	mt[#mt + 1] = {startBeat
					+ 2*duration,	"Hidden",		0.0,			0.0,		pnOther}
	mt[#mt + 1] = {startBeat
					+ 2*duration,	"Sudden",		0.0,			0.0,		pn}

	return mt
end


-- Perframe controller
_FG_[#_FG_ + 1] = Def.ActorFrame {
	InitCommand = function(self)
		self:SetUpdateFunction(gfxUpdateFunction)
	end,

	Def.ActorFrame {
		InitCommand = function(self)
			self:sleep(69420)
		end
	}
}


-- Load the mods table parser into this script.
modsTable = {
	-- [1]: beat start
	-- [2]: mod type
	-- [3]: mod strength (out of unity),
	-- [4]: mod approach (in beats to complete)
	-- [5]: player application (1 = P1, 2 = P2, 3 = both, 0 = neither)
		
		{   0.0,	"ScrollBPM",		200.0,    0.0,	3}, 
		{   0.0,	"ScrollSpeed",	niceSpeed,    0.0,	3}, 
		{   0.0,	"Bumpy",			  0.05,   0.0,	3}, 

		{   0.0,	"Dark",				  1.0,    0.0,	3},
		{   0.0,	"Flip",				  0.5,    0.0,	3},
		{  18.0,	"Dark",				  0.0,    0.0,	3},
		{  18.0,	"Flip",				  0.0,    2.0,	3},



		{  32.000,	"Cross",			  0.2,    0.25,	3},
		{  32.750,	"Invert",			  1.0,    0.25,	3},
		{  33.500,	"Cross",			  0.0,    0.25,	3},
		{  34.000,	"Split",			  0.2,    0.25,	1},
		{  34.000,	"Reverse",			  0.2,    0.25,	2},
		{  34.000,	"Split",			 -0.2,    0.25,	2},
		{  34.750,	"Invert",			  0.0,    0.25,	3},
		{  35.000,	"Reverse",			  0.0,    1.00,	3},
		{  35.000,	"Split",			  0.0,    1.00,	3},

		{ 36.000,	"ScrollSpeed",	expnSpeed,  128.00,	3},
		{ 36.000,	"Expand",			  1.0,  128.00,	3},
		{ 36.000,	"Drunk",			  2.0,  128.00,	1},
		{ 36.000,	"Drunk",			  2.0,  128.00,	2},
		{ 36.000,	"Mini",				  0.3,  128.00,	3},
		{161.000,	"Mini",				  0.0,    1.00,	3},
		{161.000,	"Drunk",			  0.0,    1.00,	3},
		{161.000,	"Expand",			  0.0,    1.00,	3},
		{161.000,	"ScrollSpeed",	niceSpeed,    1.00,	3},

		{162.000,	"Mini",				  0.4,    1.00,	3},
		{162.000,	"Wave",				  0.5,    1.00,	3},
		{162.000,	"Cross",			  1.0,    1.00,	3},
		{162.000,	"Blind",			  1.0,    1.00,	3},
		{162.000,	"Centered",			  1.0,    1.00,	3},
		{162.000,	"Reverse",			  0.1,    1.00,	1},
		{162.000,	"Reverse",			  0.9,    1.00,	2},
		{162.000,	"Xmode",			  0.6,    1.00,	1},
		{162.000,	"Xmode",			  0.6,    1.00,	2},
		{164.000,	"Drunk",			  0.5,   96.00,	3},

		{256.000,	"Mini",				  0.0,    4.00,	3},
		{256.000,	"Wave",				  0.0,    4.00,	3},
		{256.000,	"Cross",			  0.0,    1.00,	3},
		{256.000,	"Blind",			  0.0,    1.00,	3},
--		{256.000,	"Centered",			  0.0,    1.00,	3},
		{256.000,	"Reverse",			  0.0,    1.00,	3},
		{256.000,	"Xmode",			  0.0,    2.00,	3},
		{256.000,	"Drunk",			  0.0,    1.00,	3},


		{260.000,	"ScrollSpeed",	slowSpeed,    0.00,	3},
		{290.000,	"ScrollSpeed",	niceSpeed,    2.00,	3},

		{354.000,	"Wave",				  0.5,    4.00,	3},
		{356.000,	"Wave",				  0.8,   32.00,	3},
		{388.000,	"Wave",				  0.2,   28.00,	1},
		{388.000,	"Wave",				  1.5,   28.00,	2},
		{416.000,	"Wave",				  0.0,    4.00,	3},

		{380.000, 	"Dark", 			  1.0,    1.00, 3},
		{380.000, 	"Stealth", 			  0.7,    1.00, 3},
		{382.000, 	"Dark", 			  0.0,    2.00, 3},
		{382.000, 	"Stealth", 			  0.0,    2.00, 3},

		{402.000, 	"Dark", 			  1.0,    2.00, 3},
		{402.000, 	"Stealth", 			  0.7,    2.00, 3},
		{402.000, 	"Dark", 			  0.0,    2.00, 3},
		{404.000, 	"Stealth", 			  0.0,    2.00, 3},

		{450.000,	"ScrollSpeed",	slowSpeed,    2.00,	3},
		{450.000,	"Dizzy",			  1.0,    2.00,	1},
		{450.000,	"Dizzy",			 -1.0,    2.00,	2},
		{452.000,	"Dizzy",			  0.0,   16.00,	3},


		{464.000,	"Expand",			  0.3,    4.00,	3},
		{468.000,	"Expand",			  0.8,   16.00,	3},
		{484.000,	"Expand",			  0.2,   24.00,	1},
		{484.000,	"Expand",			  1.5,   24.00,	2},
		{504.000,	"Expand",			  0.0,    4.00,	3},
		{504.000,	"Tiny",				  0.5,    4.00,	3},

		{ 540.900,	"Tiny",				  0.0,    0.0,	3},
		{ 540.900,	"Expand",			  2.0,    0.0,	1}, 
		{ 540.900,	"Wave",				  2.0,    0.0,	2}, 
		{ 540.900,	"Tornado",			  1.0,    0.1,	1},
		{ 540.900,	"Tornado",			 -1.0,    0.1,	2},
		{ 540.900,	"Mini",				  0.7,    0.1,	3},
		{ 540.900,	"Tipsy",			  1.0,    0.1,	3},
		{ 540.900,	"Bumpy",		     10.0,    0.1,	3},
		{ 541.000,	"Tipsy",			  0.0,    7.0,	3},
		{ 541.000,	"Mini",				  0.0,    7.0,	3},
		{ 541.000,	"Tornado",			  0.0,    7.0,	3},
		{ 541.000,	"Bumpy",		      0.0,    7.0,	3},
}

function AddURShift(playerNumber, beatIndex, duration, shiftU, shiftR)
	local i =  0.75 * shiftU - 0.25 * shiftR
	local f = -0.25 * shiftU - 0.25 * shiftR

	modsTable[#modsTable + 1] = {beatIndex,	"Flip",     f,	duration,	playerNumber}
	modsTable[#modsTable + 1] = {beatIndex,	"Invert",   i,	duration,	playerNumber}
end

modsTable = EchoRhythmAddMods(modsTable, {2,  35.750, 8})
modsTable = EchoRhythmAddMods(modsTable, {2,  55.750, 4})
modsTable = EchoRhythmAddMods(modsTable, {2,  67.750, 8})
modsTable = EchoRhythmAddMods(modsTable, {2,  87.750, 4})
modsTable = EchoRhythmAddMods(modsTable, {1, 101.250, 8})
modsTable = EchoRhythmAddMods(modsTable, {1, 121.250, 4})
modsTable = EchoRhythmAddMods(modsTable, {1, 131.750, 8})
modsTable = EchoRhythmAddMods(modsTable, {1, 151.750, 4})

modsTable = EchoRhythmAddMods(modsTable, {2, 260,     8})
modsTable = EchoRhythmAddMods(modsTable, {1, 280,     4})

-- modsTable = EchoRhythmAddMods(modsTable, {2, 378,     2})
-- modsTable = EchoRhythmAddMods(modsTable, {1, 402,     2})

for beatIndex = 260,290,2 do
	modsTable[#modsTable+1] = {beatIndex,			"Centered",	  0.8 * slowSpeed,    2.00,	3}
	modsTable[#modsTable+1] = {beatIndex,			"Flip",		  			-0.25,    2.00,	3}
	modsTable[#modsTable+1] = {beatIndex + 1.95,	"Centered",				  0.0,    0.00,	3}
	modsTable[#modsTable+1] = {beatIndex + 1.95,	"Flip",					  0.0,    0.00,	3}
end

scrapeLaneChanges = {
	-- [1] Start beat
	-- [2] Duration in
	-- [3] Duration hold
	-- [4] Duration out
	-- [5] Flip
	-- [6] Invert
	{288.00, 0.75, 0.00, 1.25, -2, -2},
	{290.00, 0.75, 0.00, 1.25, -1, -1},
	{304.00, 0.75, 0.00, 1.25,  1,  1},
	{306.00, 0.75, 0.00, 1.25,  1, -1},
	{320.00, 0.75, 0.00, 1.25, -1,  0},
		{322.25, 0.125, 0.375, 0.00, -1, -1},
		{322.75, 0.125, 0.375, 0.00, -2, -2},
		{323.25, 0.125, 0.375, 0.00,  0,  0},	-- reset

	{336.00, 0.75, 0.00, 1.25,  2,  0},
	{338.00, 0.75, 0.00, 1.25, -3,  0},
		{352.25, 0.125, 0.375, 0.00, -1, -1},
		{352.75, 0.125, 0.375, 0.00, -2, -2},
		{353.25, 0.125, 0.375, 0.00,  0,  0},	-- reset

		{354.25, 0.125, 0.375, 0.00, -1, -1},
		{354.75, 0.125, 0.375, 0.00, -2, -2},
		{355.25, 0.125, 0.375, 0.00,  0,  0},	-- reset

	{384.00, 0.75, 0.00, 1.25, -1, -1},
	{386.00, 0.75, 0.00, 1.25, -2, -2},
	{416.00, 0.75, 0.00, 1.25, -1, -1},
	{418.00, 0.75, 0.00, 1.25, -1, -2},

	{480.00, 0.75, 0.00, 1.25, -1,  0},
	{482.00, 0.75, 0.00, 1.25, -1,  0},
}

for beatIndex = 0,3 do
	local trueBeat = 292 + beatIndex * 16
	for subBeatIndex = 0,10,2 do
		local sbs = (subBeatIndex % 4 == 2) and -1 or 1
		modsTable[#modsTable+1] = {trueBeat + subBeatIndex,			"Tipsy",	  0.8 * sbs,    0.00,	3}
		modsTable[#modsTable+1] = {trueBeat + subBeatIndex,			"Drunk",	  0.8 * sbs,    0.00,	3}
		modsTable[#modsTable+1] = {trueBeat + subBeatIndex,			"Stealth",	  0.4 * sbs,    0.00,	3}
		modsTable[#modsTable+1] = {trueBeat + subBeatIndex + 0.05,	"Tipsy",	  0.0,		    2.00,	3}
		modsTable[#modsTable+1] = {trueBeat + subBeatIndex + 0.05,	"Drunk",	  0.0,		    2.00,	3}
		modsTable[#modsTable+1] = {trueBeat + subBeatIndex + 0.05,	"Stealth",	  0.0,		    2.00,	3}
	end
end

for scrapeIndex = 1,#scrapeLaneChanges do
	local sbt = scrapeLaneChanges[scrapeIndex][1]
	local sdi = scrapeLaneChanges[scrapeIndex][2]
	local sdh = scrapeLaneChanges[scrapeIndex][3]
	local sdo = scrapeLaneChanges[scrapeIndex][4]
	local sus = scrapeLaneChanges[scrapeIndex][5]
	local srs = scrapeLaneChanges[scrapeIndex][6]

	AddURShift(3, sbt, sdi, sus, srs)
	modsTable[#modsTable+1] = {sbt,			"Dark",		  0.95,   sdi,	3}
	if sdo > 0 then		
		AddURShift(3, sbt+sdi+sdh, sdo, 0, 0)
	end
	modsTable[#modsTable+1] = {sbt+sdi+sdh,	"Dark",		  0.0,    sdo,	3}

	if sdh == 0 then
		modsTable[#modsTable+1] = {sbt,				"Mini",		  0.4,	    sdi,	3}
		modsTable[#modsTable+1] = {sbt+sdi,			"Mini",		 -0.4,	    0.0,	3}
		modsTable[#modsTable+1] = {sbt+sdi+0.05,	"Mini",		  0.0,	    sdi,	3}
	end
end


for tripletingBeat = 508,524,16 do
	modsTable[#modsTable+1] = {tripletingBeat,			"Flip",		  1.0,	  8.0,	3}
	modsTable[#modsTable+1] = {tripletingBeat + 4.00,	"Flip",		  0.0,	  8.0,	3}
end


modsTable = SortModsTable(modsTable)
_FG_[#_FG_ + 1] = LoadActor("./modsHQ.lua", {modsTable, 0})



return _FG_
